import os

# Get the path of the current Python module
module_path = os.path.abspath(__file__)

# Construct the path to the file you need to open
file_path = os.path.join(os.path.dirname(module_path), "x_range_change_callback_code.js")

# Open the file and read its contents
with open(file_path, 'r') as handle:
    x_range_change_callback_code = ''.join(handle.readlines())

# Construct the path to the file you need to open
file_path = os.path.join(os.path.dirname(module_path), "search_callback_code.js")

with open(file_path, 'r') as handle:
    search_callback_code =''.join(handle.readlines())

def reload():
    pass


# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_browser.ipynb.

# %% auto 0
__all__ = ['GenomeBrowser', 'Track']

# %% ../nbs/00_browser.ipynb 4
from fastcore.basics import *

from genomenotebook.plotting import (
    get_genome_annotations,
    get_gene_patches, 
    get_genes_from_annotation, 
    create_genome_browser_plot,
    get_all_glyphs)
from bokeh.models import (
    CustomJS,
    Range1d,
    ColumnDataSource,
    LabelSet, 
    TextInput,
    Rect,
    Div,
    Styles,
    NumeralTickFormatter
)
from bokeh.plotting import show, figure
from bokeh.layouts import column, row

from Bio import SeqIO
from .js_callback_code import x_range_change_callback_code, search_callback_code
from bokeh.io import output_notebook
import pyBigWig
import numpy as np
import pandas as pd
output_notebook()


# %% ../nbs/00_browser.ipynb 5
class GenomeBrowser:
    def __init__(self,
                 genome_path: str, #path to the fasta file of the genome sequence
                 gff_path: str, #path to the gff3 file of the annotations
                 init_pos=None, #initial position to display
                 init_win=10000, #initial window size (max=20000)
                 bounds=None, #bounds can be specified. This helps preserve memory by not loading the whole genome if not needed.
                 show_seq=True, #shows the sequence when zooming in
                 search=True, #enables a search bar to lookup a gene name or a DNA sequence
                 **kwargs):
        self.genome_path=genome_path
        genome = next(SeqIO.parse(self.genome_path, 'fasta'))
        self.genome_id = genome.id
        self.genome_size=len(genome.seq)
        
        if not bounds:
            self.bounds=(0,self.genome_size)
        else:
            self.bounds=bounds
            
        self.seq=genome.seq[self.bounds[0]:self.bounds[1]]
   
        self.gff_path=gff_path

        if not init_pos:
            self.init_pos=sum(self.bounds)//2
        elif init_pos>bounds[1] or init_pos<bounds[0]:
            print("WARNING: Requested an initial position outside of the browser bounds")
            self.init_pos=sum(self.bounds)//2
        else:
            self.init_pos=init_pos

        self.init_win=min(init_win,self.bounds[1]-self.bounds[0])

        self.show_seq=show_seq
        self.max_glyph_loading_range=20000
        self.frame_width=600

        

        self.elements=self.get_browser(show_seq=show_seq)

        if search:
            self.search=self.get_search_box()
            self.elements=[self.search]+self.elements
        else:
            self.search=None
            
        self.tracks=[]


    def get_browser(self,show_seq=True):

        annotation = get_genome_annotations(self.gff_path,self.bounds)
        annotation.gene = annotation.gene.fillna("")
        annotation.gene = annotation[['gene', 'locus_tag']].apply(lambda x: f"{x[0]} {x[1]}", axis=1)
        genes = get_genes_from_annotation(annotation)

        semi_win = self.init_win / 2
        x_range = Range1d(
            max(self.bounds[0],self.init_pos - semi_win), min(self.bounds[1],self.init_pos + semi_win), 
            bounds=self.bounds, 
            max_interval=100000,
            min_interval=40
        )

        self.glyph_source = ColumnDataSource(get_gene_patches(genes, x_range.start, x_range.end))

        #This will contain the glyphs plotted by bokeh
        loaded_glyph_source = ColumnDataSource({"start":[x_range.start],"end":[x_range.end], "range":[self.max_glyph_loading_range]})

        #This contains the glyphs for the whole genome
        self.all_glyphs=get_all_glyphs(genes, self.bounds)

        p = create_genome_browser_plot(self.glyph_source, x_range, output_backend="webgl")
        p.frame_width=self.frame_width

        sty=Styles(font_size='14px',
                font_family="Courrier",
                color="black",
                display="inline-block",
                background_color = "white",
                margin="0",
                margin_left= "2px",
                )
        
        ## Adding the ability to display the sequence when zooming in
        sequence = {
            'seq': str(self.seq).upper(),
            'bounds':self.bounds
        }

        self.div = Div(height=18, height_policy="fixed", 
                    width=600, width_policy="fixed",
                    styles = sty
                    )

        xcb = CustomJS(
            args={
                "x_range": p.x_range,
                "sequence": sequence,
                "all_glyphs":self.all_glyphs,
                "glyph_source": self.glyph_source,
                "div": self.div,
                "loaded_glyph_source":loaded_glyph_source,
            },
            code=x_range_change_callback_code
        )

        p.x_range.js_on_change('start', xcb)
        self.x_range=p.x_range
        self.gene_track=p

        if self.show_seq:
            return [p,self.div]
        else:
            return [p]

    def get_search_box(self):

        ## Create a text input widget for search
        text_input = TextInput(value="")

        ## Adding BoxAnnotation to highlight search results
        search_span_source = ColumnDataSource({"x":[],"width":[]})#"y":[]
        h=Rect(x='x',y=-2,width='width',height=self.gene_track.height,fill_color='green',fill_alpha=0.2,line_alpha=0)
        self.gene_track.add_glyph(search_span_source, h)

        ## Adding the ability to display the sequence when zooming in
        sequence = {
            'seq': str(self.seq).upper(),
            'bounds':self.bounds
        }

        call_back_search = CustomJS(
            args={
                "x_range": self.x_range,
                "glyph_source": self.glyph_source,
                "sequence": sequence,
                "all_glyphs": self.all_glyphs,
                "text_input": text_input,
                "search_span_source": search_span_source,
                "div": self.div,
            },
            code=search_callback_code
        )

        text_input.js_on_change('value',call_back_search)#,xcb)

        return text_input
    
    def show(self):
        show(column(self.elements + [t.fig for t in self.tracks]))



# %% ../nbs/00_browser.ipynb 7
class Track:
    def __init__(self,
                 height: int = 200, #size of the track
                ):
        self.height = height
        self.fig = figure(tools="xwheel_zoom,xpan,save,reset",
                          active_scroll="xwheel_zoom",
                          height=height,
                          y_axis_location="right", #this is required in order to keep a proper alignment with the sequence
                          output_backend="webgl")
        self.fig.xaxis[0].formatter = NumeralTickFormatter(format="0,0")
        


# %% ../nbs/00_browser.ipynb 9
@patch
def add_track(self:GenomeBrowser,
             height:int = 200 #size of the track
             ) -> Track:
    """Adds a track to the the GenomeBrowser. Ensures that the x_range are shared and figure widths are identical."""
    t = Track(height=height)
    t.fig.x_range = self.x_range
    t.fig.frame_width = self.frame_width
    t.bounds = self.bounds
    self.tracks.append(t)
    return t
    

# %% ../nbs/00_browser.ipynb 13
@patch
def line(self:Track,
         source: pd.DataFrame, #pandas DataFrame containing the data
         pos: str, #name of the column containing the positions along the genome
         y: str, #name of the column containing the data to be plotted on the y-axis
        ):
    source=source.loc[(self.bounds[0] < source[pos]) & (source[pos] < self.bounds[1])]
    self.fig.line(source=source, x=pos, y=y)


# %% ../nbs/00_browser.ipynb 17
from bokeh.transform import linear_cmap, factor_cmap

# %% ../nbs/00_browser.ipynb 18
@patch
def scatter(self:Track,
         source: pd.DataFrame, #pandas DataFrame containing the data
         pos: str, #name of the column containing the positions along the genome
         y: str, #name of the column containing the data to be plotted on the y-axis
         z: str = None, #name of a column containing numerical data rendered as a linear color map (cannot be used for line plots)
         factors: str = None, #name of a column of values to be used as factors
        ):
    source=source.loc[(self.bounds[0] < source[pos]) & (source[pos] < self.bounds[1])]
    
    if factors!=None:
        color=factor_cmap(factors,"Category10_3",tuple(set(source[factors].values)))
        
        self.fig.scatter(source=source, x=pos, y=y, color=color, legend_group=factors)

        self.fig.legend.location = "top_left"
        self.fig.legend.title = "ori"
    elif z!=None:
        pass
    else:
        self.fig.scatter(source=source, x=pos, y=y)


# %% ../nbs/00_browser.ipynb 21
@patch
def bar(self:Track,
         source: pd.DataFrame, #pandas DataFrame containing the data
         pos: str, #name of the column containing the positions along the genome
         y: str, #name of the column containing the data to be plotted on the y-axis
         z: str = None, #name of a column containing numerical data rendered as a linear color map (cannot be used for line plots)
         factors: str = None, #name of a column of values to be used as factors
        ):
    source=source.loc[(self.bounds[0] < source[pos]) & (source[pos] < self.bounds[1])]
    
    if factors!=None:
        color=factor_cmap(factors,"Category10_3",tuple(set(source[factors].values)))
        
        self.fig.vbar(source=source, x=pos, top=y, color=color, legend_group=factors)

        self.fig.legend.location = "top_left"
        self.fig.legend.title = "ori"
    elif z!=None:
        pass
    else:
        self.fig.vbar(source=source, x=pos, top=y)
        



# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/genomenotebook',
                'doc_host': 'https://dbikard.github.io',
                'git_url': 'https://github.com/dbikard/genomenotebook',
                'lib_path': 'genomenotebook'},
  'syms': { 'genomenotebook.browser': { 'genomenotebook.browser.GenomeBrowser': ('browser.html#genomebrowser', 'genomenotebook/browser.py'),
                                        'genomenotebook.browser.GenomeBrowser.__init__': ( 'browser.html#genomebrowser.__init__',
                                                                                           'genomenotebook/browser.py'),
                                        'genomenotebook.browser.GenomeBrowser.add_track': ( 'browser.html#genomebrowser.add_track',
                                                                                            'genomenotebook/browser.py'),
                                        'genomenotebook.browser.GenomeBrowser.get_browser': ( 'browser.html#genomebrowser.get_browser',
                                                                                              'genomenotebook/browser.py'),
                                        'genomenotebook.browser.GenomeBrowser.get_search_box': ( 'browser.html#genomebrowser.get_search_box',
                                                                                                 'genomenotebook/browser.py'),
                                        'genomenotebook.browser.GenomeBrowser.show': ( 'browser.html#genomebrowser.show',
                                                                                       'genomenotebook/browser.py'),
                                        'genomenotebook.browser.Track': ('browser.html#track', 'genomenotebook/browser.py'),
                                        'genomenotebook.browser.Track.__init__': ( 'browser.html#track.__init__',
                                                                                   'genomenotebook/browser.py'),
                                        'genomenotebook.browser.Track.bar': ('browser.html#track.bar', 'genomenotebook/browser.py'),
                                        'genomenotebook.browser.Track.line': ('browser.html#track.line', 'genomenotebook/browser.py'),
                                        'genomenotebook.browser.Track.scatter': ( 'browser.html#track.scatter',
                                                                                  'genomenotebook/browser.py')},
            'genomenotebook.examples': {},
            'genomenotebook.js_callback_code': {},
            'genomenotebook.plotting': { 'genomenotebook.plotting.arrow_patch': ('plotting.html#arrow_patch', 'genomenotebook/plotting.py'),
                                         'genomenotebook.plotting.create_genome_browser_plot': ( 'plotting.html#create_genome_browser_plot',
                                                                                                 'genomenotebook/plotting.py'),
                                         'genomenotebook.plotting.get_all_glyphs': ( 'plotting.html#get_all_glyphs',
                                                                                     'genomenotebook/plotting.py'),
                                         'genomenotebook.plotting.get_arrow_patch': ( 'plotting.html#get_arrow_patch',
                                                                                      'genomenotebook/plotting.py'),
                                         'genomenotebook.plotting.get_gene_patches': ( 'plotting.html#get_gene_patches',
                                                                                       'genomenotebook/plotting.py'),
                                         'genomenotebook.plotting.get_genes_from_annotation': ( 'plotting.html#get_genes_from_annotation',
                                                                                                'genomenotebook/plotting.py'),
                                         'genomenotebook.plotting.get_genome_annotations': ( 'plotting.html#get_genome_annotations',
                                                                                             'genomenotebook/plotting.py'),
                                         'genomenotebook.plotting.get_y_range': ('plotting.html#get_y_range', 'genomenotebook/plotting.py'),
                                         'genomenotebook.plotting.rect_patch': ('plotting.html#rect_patch', 'genomenotebook/plotting.py')}}}



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/01_plotting.ipynb.

# %% auto 0
__all__ = ['Y_RANGE', 'gene_y_range', 'create_genome_browser_plot', 'get_gene_patches', 'get_y_range', 'get_genome_annotations',
           'get_genes_from_annotation', 'get_all_glyphs', 'rect_patch', 'arrow_patch', 'get_arrow_patch']

# %% ../nbs/01_plotting.ipynb 4
import gffpandas.gffpandas as gffpd
import numpy as np

from bokeh.plotting import figure
from bokeh.models.tools import BoxZoomTool
from bokeh.models import HoverTool, NumeralTickFormatter, LabelSet
from bokeh.models.glyphs import Patches
from bokeh.models import (
    CustomJS,
    Range1d,
    ColumnDataSource,
)

# %% ../nbs/01_plotting.ipynb 5
def create_genome_browser_plot(glyphSource, x_range, **kwargs):
    plot_height = kwargs.get("plot_height", 150)
    label_angle = kwargs.get("label_angle", 45)
    text_font_size = kwargs.get("text_font_size", "10pt")
    output_backend = kwargs.get("output_backend", "webgl")
    
    y_min, y_max = get_y_range()
    p_annot = figure(
        tools = "xwheel_zoom,xpan,save",
        active_scroll = "xwheel_zoom",
        height = plot_height,
        x_range = x_range,
        y_range = Range1d(y_min, y_max),
        output_backend=output_backend,
    )
    # Add tool
    p_annot.add_tools(BoxZoomTool(dimensions="width"))

    #p_annot.sizing_mode = "stretch_both"

    # Format x axis values
    p_annot.xaxis[0].formatter = NumeralTickFormatter(format="0,0")
    # Hide grid
    p_annot.xgrid.visible = False
    p_annot.ygrid.visible = False
    # Hide axis
    p_annot.yaxis.visible = False
    glyph = p_annot.add_glyph(
        glyphSource, Patches(xs="xs", ys="ys", fill_color="color")
    )
    # gene labels in the annotation track
    # This seems to be necessary to show the labels
    p_annot.scatter(x="pos", y=0, size=0, source=glyphSource)
    labels = LabelSet(
        x="pos",
        y=-0.9,
        text="names",
        level="glyph",
        angle=label_angle,
        text_font_size=text_font_size,
        x_offset=-5,
        y_offset=0,
        source=glyphSource,
        text_align='left',
    )

    p_annot.add_layout(labels)
    p_annot.add_tools(
        HoverTool(
            renderers=[glyph],
            tooltips=[("name", "@hover_names"), ("product", "@product")],
        )
    )
    return p_annot

# %% ../nbs/01_plotting.ipynb 6
def get_gene_patches(genes, left, right):
    genes_region = genes[
        (genes["right"] > left)
        & (genes["left"] < right)
        & (genes["type"] != "repeat_region")
    ]
    arr = arrow_patch(genes_region)
    # repeat_region
    rep_region = genes[
        (genes["right"] > left)
        & (genes["left"] < right)
        & (genes["type"] == "repeat_region")
    ]
    rect = rect_patch(rep_region)

    # concatenate patches
    res = dict([(k, arr[k] + rect[k]) for k in arr.keys()])
    return res


Y_RANGE = (-2, 2)
def get_y_range():
    """Accessor that returns the Y range for the genome browser plot

    :return: [description]
    :rtype: [type]
    """
    return Y_RANGE


def get_genome_annotations(genome_path: str, bounds=None):
    annotation = gffpd.read_gff3(genome_path)
    annotation = annotation.attributes_to_columns()
    if bounds:
        annotation = annotation.loc[(annotation.start<bounds[1]) & (annotation.end>bounds[0])]

    annotation.loc[:, "left"] = annotation[["start"]].values
    annotation.loc[:, "right"] = annotation[["end"]].values
    return annotation

def get_genes_from_annotation(annotation):

    genes = annotation[
        annotation.type.isin(["CDS", "repeat_region", "ncRNA", "rRNA", "tRNA"])
    ].copy()

    genes.loc[genes["strand"] == "+", "start"] = genes.loc[
        genes["strand"] == "+", "left"
    ].values

    genes.loc[genes["strand"] == "+", "end"] = genes.loc[
        genes["strand"] == "+", "right"
    ].values

    genes.loc[genes["strand"] == "-", "start"] = genes.loc[
        genes["strand"] == "-", "right"
    ].values

    genes.loc[genes["strand"] == "-", "end"] = genes.loc[
        genes["strand"] == "-", "left"
    ].values

    genes.loc[genes["type"] == "repeat_region", "gene"] = "REP"
    return genes

def get_all_glyphs(genes,bounds:tuple):
    all_glyphs=get_gene_patches(genes, bounds[0], bounds[1])

    ks=list(all_glyphs.keys())
    ref_list_ix=ks.index('xs')
    # Sort all the lists in the dictionary based on the values of the reference list
    sorted_lists = sorted(zip(*[all_glyphs[k] for k in ks]), key= lambda x: x[ref_list_ix][0])

    # Convert the sorted tuples back into separate lists
    unzipped_lists = zip(*sorted_lists)

    # Create a new dictionary with the same keys as the original dictionary, but with the sorted lists as values
    all_glyphs = {k: list(t) for k, t in zip(ks, unzipped_lists)}
    
    return all_glyphs

# %% ../nbs/01_plotting.ipynb 7
def rect_patch(genes_region):
    y_min, y_max = gene_y_range
    xs = list(
        zip(
            genes_region.start.values,
            genes_region.start.values,
            genes_region.end.values,
            genes_region.end.values,
        )
    )
    xs = [np.array(x) for x in xs]
    ys = [np.array([y_min, y_max, y_max, y_min]) for i in range(genes_region.shape[0])]
    genes_mid = genes_region.left + (genes_region.right - genes_region.left) / 2
    pos = list(genes_mid.values)
    names = list(genes_region.gene.values)
    product = list(genes_region["product"].values)
    color = ["grey"] * genes_region.shape[0]
    return dict(
        xs=xs,
        ys=ys,
        pos=pos,
        names=[""] * genes_region.shape[0],
        hover_names=names,
        product=product,
        color=color,
    )

# %% ../nbs/01_plotting.ipynb 8
def arrow_patch(genes_region):
    arr_plus = get_arrow_patch(genes_region[genes_region["strand"] == "+"], "+")
    arr_minus = get_arrow_patch(genes_region[genes_region["strand"] == "-"], "-")
    return dict([(k, arr_plus[k] + arr_minus[k]) for k in arr_plus.keys()])

# %% ../nbs/01_plotting.ipynb 9
gene_y_range = (-1.5, -1)

def get_arrow_patch(genes_region, ori="+"):
    y_min, y_max = gene_y_range
    y_min = y_min 
    if ori == "+":
        xs = list(
            zip(
                genes_region.start.values,
                genes_region.start.values,
                np.maximum(genes_region.start.values, genes_region.end.values - 100),
                genes_region.end.values,
                np.maximum(genes_region.start.values, genes_region.end.values - 100),
            )
        )
        color = ["orange"] * genes_region.shape[0]
    elif ori == "-":
        xs = list(
            zip(
                genes_region.start.values,
                genes_region.start.values,
                np.minimum(genes_region.start.values, genes_region.end.values + 100),
                genes_region.end.values,
                np.minimum(genes_region.start.values, genes_region.end.values + 100),
            )
        )
        color = ["purple"] * genes_region.shape[0]

    ys = [
        np.array([y_min, y_max, y_max, (y_max + y_min) / 2, y_min])
        for i in range(genes_region.shape[0])
    ]
    genes_mid = genes_region.left + (genes_region.right - genes_region.left) / 2
    pos = list(genes_mid.values)
    names = list(genes_region.gene.values)
    product = list(genes_region["product"].values)
    return dict(
        xs=xs,
        ys=ys,
        pos=pos,
        names=names,
        hover_names=names,
        product=product,
        color=color,
    )
